/* thing to do
1)* auto backup
2)* undo
3)* redo
4)* back/forward across file boundaries
5)* open/save playlist files (with relative paths if possible to improve portability)
6)NO open/save group files - probably not, if I save play position in playlist file
7)* loop if loop file exists, loop the loop, otherwise loop the file
8) write a help file
9)* create a program icon
10)* save/restore volume setting in settings ini file
11) compile it on Linux
12)* allow drag/drop playlist files to playlist (or main window)
13)* allow drag/drop folders to playlist (this is just a convenience)
14)* use layout on main window for resizing and stretching
15)* keyboard shortcut to toggle loop mode
 */
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "breaktimer.h"
#include "loop.h"
#include "settings.h"
#include "shortcuts.h"
#include <QMessageBox>
#include <QSettings>
#include <QDir>
#include <QDesktopServices>
#include <QUrl>
#include <QTimer>
#include <QSound>
#include <phonon>
#include <QFileDialog>
#include <QCloseEvent>
#include <QProcess>
/*
A few things to be aware of when using the phonon media classes
1) The aboutToFinish signal is emitted shortly before the current file
   finishes playing. This is used to enqueue the next file to play before
   the current file finished playing. However, if a seek backwards is done
   after enqueueing the next file, but before the current file finishes,
   another aboutToFinish will be emitted when the current file is once
   again about to finish playing. If the next file is enqueued again,
   it will then play twice! The solution is to clear the queue before
   enqueueing the next file when handling the aboutToFinish signal.
2) When seeking to a position in a new file (after setCurrentSource())
   I need to wait until the current source changes and enters the paused
   or play state before I can seek to the desired position.
*/
/*
To create a proper Windows Icon file for the Windows EXE file:
First, create an ICO format bitmap file that contains the icon image.
This can be done with e.g. Microsoft Visual C++:
Select File|New, then select the File tab in the dialog that appears, and choose Icon.
(Note that you do not need to load your application into Visual C++; here we are only using the icon editor.)
Or use GIMP to create the ICO file!
Store the ICO file in your application's source code directory,
for example, with the name myappico.ico.
Then, create a text file called, say, myapp.rc in which you put a single line of text:

    IDI_ICON1               ICON    DISCARDABLE     "myappico.ico"

Finally, assuming you are using qmake to generate your makefiles, add this line to your myapp.pro file:

    RC_FILE = myapp.rc

Regenerate your makefile and your application. The .exe file will now be represented with your icon in Explorer.
*/

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    setCurrentPath();
    // set up the QMap of action names to slots (12 actions for user shortcut keys)
    shortcutSlots["PlayPause"]=SLOT(playpause());
    shortcutSlots["Zero"]=SLOT(zero());
    shortcutSlots["Rewind"]=SLOT(rewind());
    shortcutSlots["Back"]=SLOT(back());
    shortcutSlots["Forward"]=SLOT(forward());
    shortcutSlots["Timer"]=SLOT(tiptimer());
    shortcutSlots["VolumeUp"]=SLOT(volup());
    shortcutSlots["VolumeDown"]=SLOT(voldown());
    shortcutSlots["NextTrack"]=SLOT(next());
    shortcutSlots["PrevTrack"]=SLOT(prev());
    shortcutSlots["Undo"]=SLOT(undo());
    shortcutSlots["Redo"]=SLOT(redo());
    shortcutSlots["Loop"]=SLOT(loop());
    shortcutSlots["BreakTimer"]=SLOT(dobreaktimer());
    emptyUndo();
    queuedFlag=false;
    // create and link the phonon audio output and media objects
    audioOutput = new Phonon::AudioOutput(Phonon::MusicCategory, this);
    mediaObject = new Phonon::MediaObject(this);
    Phonon::createPath(mediaObject, audioOutput);
    // link the seek and volume sliders to the media object and audio output
    ui->seekSlider->setMediaObject(mediaObject);
    ui->volumeSlider->setAudioOutput(audioOutput);
    // connect the metadata changed and time tick signals to my slots
    connect(mediaObject,SIGNAL(metaDataChanged()),this,SLOT(metaDataChanged()));
    mediaObject->setTickInterval(1000);
    connect(mediaObject,SIGNAL(tick(qint64)),this,SLOT(tick(qint64)));
    connect(mediaObject,SIGNAL(aboutToFinish()),this,SLOT(enqueueNext()));
    connect(mediaObject,SIGNAL(currentSourceChanged(Phonon::MediaSource)),this,SLOT(sourceChanged()));
    connect(mediaObject,SIGNAL(stateChanged(Phonon::State,Phonon::State)),this,SLOT(stateChanged(Phonon::State,Phonon::State)));
    connect(mediaObject,SIGNAL(totalTimeChanged(qint64)),this,SLOT(totalTimeChanged(qint64)));
    connect(mediaObject,SIGNAL(prefinishMarkReached(qint32)),this,SLOT(prefinishMark()));
    // set up an empty playlist model
    QStringList list;
    playlistModel = new PlaylistModel(list,this);
    ui->listView_playlist->setModel(playlistModel);
    // make del key a shortcut for deleting entries from playlist
    QAction *del = new QAction(this);
    del->setShortcut(Qt::Key_Delete);
    del->setShortcutContext(Qt::WidgetShortcut); // shortcut enabled only when playlist has focus
    ui->listView_playlist->addAction(del); // add action to the playlist view widget
    connect(del,SIGNAL(triggered()),this,SLOT(deletePlaylistItems()));
    // make double-click a playlist entry play the file
    connect(ui->listView_playlist,SIGNAL(doubleClicked(QModelIndex)),this,SLOT(DoubleclickPlaylist(QModelIndex)));
    // link signals/slots for menus, pushbuttons etc.
    connect(ui->actionOptions,SIGNAL(triggered()),this,SLOT(dosettings()));
    connect(ui->actionShortcuts,SIGNAL(triggered()),this,SLOT(doshortcuts()));
    connect(ui->actionLoop,SIGNAL(triggered()),this,SLOT(doloop()));
    connect(ui->actionBreaktimer,SIGNAL(triggered()),this,SLOT(dobreaktimer()));
    connect(ui->actionAbout,SIGNAL(triggered()),this,SLOT(about()));
    connect(ui->actionHelp,SIGNAL(triggered()),this,SLOT(help()));
    connect(ui->pushButton_playpause,SIGNAL(clicked()),this,SLOT(playpause())); // playpause
    connect(ui->pushButton_back,SIGNAL(clicked()),this,SLOT(back())); // back
    connect(ui->pushButton_forward,SIGNAL(clicked()),this,SLOT(forward())); // forward
    connect(ui->pushButton_counter,SIGNAL(clicked()),this,SLOT(zero())); // zero
    connect(ui->pushButton_rewind,SIGNAL(clicked()),this,SLOT(rewind())); // rewind
    connect(ui->pushButton_timer,SIGNAL(clicked()),this,SLOT(tiptimer())); // tip timer
    connect(ui->pushButton_next,SIGNAL(clicked()),this,SLOT(next())); // next
    connect(ui->pushButton_prev,SIGNAL(clicked()),this,SLOT(prev())); // prev
    // volup (no ui button for this, only shortcut key)
    // voldown (no ui button for this, only shortcut key)
    connect(ui->actionUndo,SIGNAL(triggered()),this,SLOT(undo())); // undo
    connect(ui->actionRedo,SIGNAL(triggered()),this,SLOT(redo())); // redo
    readSettings(); // must read the settings
    activateShortcuts(); // before activating shortcuts (shortcuts are is settings!)
    QTimer *timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(onesec()));
    timer->start(1000); // start executing onesec() every second
    TimerRunning=false; // timer is stopped
    tiptimer(); // reset the tip timer
    connect(playlistModel, SIGNAL(playlistChanged()), this, SLOT(playlistChanged()));
    connect(playlistModel,SIGNAL(playlistDropped(QString)),this,SLOT(playlistDropped(QString)));
    connect(ui->checkBox_loop, SIGNAL(stateChanged(int)), this, SLOT(loopStateChanged(int)));
    connect(ui->actionSave_as, SIGNAL(triggered()), this, SLOT(saveAs()));
    connect(ui->actionSave, SIGNAL(triggered()), this, SLOT(save()));
    connect(ui->actionClose, SIGNAL(triggered()), this, SLOT(closefile()));
    connect(ui->actionOpen_Playlist, SIGNAL(triggered()), this, SLOT(open()));
    connect(ui->actionExit, SIGNAL(triggered()), this, SLOT(close()));
    PlaylistChanged=false;
    playlistModel->setFileTypes(readFileTypes());
    openPlaylist(Openfile); // this opens the last playlist (if any)
    // show available effects for testing only
#if 0
    QList<Phonon::EffectDescription> effectDescriptions = Phonon::BackendCapabilities::availableAudioEffects();
    foreach (Phonon::EffectDescription ed, effectDescriptions) {
        QMessageBox::information(this,"Effect",QString("name:%1 desc:%2").arg(ed.name()).arg(ed.description()));
    }
#endif
#if 0
    QList<Phonon::AudioOutputDevice> audioOutputDevices = Phonon::BackendCapabilities::availableAudioOutputDevices();
    foreach (Phonon::AudioOutputDevice ad, audioOutputDevices) {
        QMessageBox::information(this,"Device",QString("name:%1 desc:%2").arg(ad.name()).arg(ad.description()));
    }
#endif
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

// set current path to path of executable file
void MainWindow::setCurrentPath(void)
{
    QStringList cla;  // command line arguments

    cla=QCoreApplication::arguments();    // get list of command line arguments (first one is application exe file)
    if (cla.size()>=1) {
          QDir::setCurrent(QFileInfo(cla[0]).absolutePath()); // set current directory to path of executable
    }
}

void MainWindow::readSettings()
{
    int size,i;
    struct shortcut newshortcut;

    QSettings settings(QSettings::IniFormat,QSettings::UserScope,"krubow","sdplayer"); // constructor, force INI file storage
    Tiptime=settings.value("tiptime",10).toInt();
    Backtime=settings.value("backtime",1).toInt();
    Skiptime=settings.value("skiptime",5).toInt();
    Volumestep=settings.value("volumestep",10).toInt();
    Soundfile=settings.value("soundfile").toString();
    PlayUsing=settings.value("playusing",0).toInt();
    Command=settings.value("command").toString();
    Openfile=settings.value("openfile").toString();
    audioOutput->setVolume(settings.value("volume",1.0).toDouble());
    size=settings.beginReadArray("shortcuts");
    for (i=0; i<size; i++) {
        settings.setArrayIndex(i);
        newshortcut.Key=settings.value("key").toString();
        newshortcut.Action=settings.value("action").toString();
        shortcutlist.append(newshortcut);
    }
    settings.endArray();
}

QStringList MainWindow::readFileTypes(void)
{
    QStringList FileTypes;
    QFile typesfile;

    FileTypes << "mp3";
    typesfile.setFileName("filetypes.txt");
    if (typesfile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QTextStream types(&typesfile);
        while (!types.atEnd()) {
            FileTypes << types.readLine().toLower();
        }
    }
    FileTypes.removeDuplicates();
    return FileTypes; // return list of supported file types from filetypes.txt ("mp3" is ALWAYS included)
}

void MainWindow::about(void)
{
    QMessageBox::about(this,"About SdPlayer","<h1>SdPlayer 1.00</h1><h2>by Keith Rubow</h2><a href=\"http://www.krubow.com/\">www.krubow.com</a>");
}

void MainWindow::help(void)
{
    QDesktopServices::openUrl(QUrl("sdplayer.html"));
}

QString formatTime(qint64 time)
{
    int minutes,seconds;

    if (time<0) {
        return "-"+formatTime(-time);
    }
    minutes=time/60000;
    seconds=(time-minutes*60000)/1000;
    return QString("%1:%2").arg(minutes).arg(seconds,2,10,QChar('0'));
}

void MainWindow::onesec(void)
{
    if (TimerRunning && Timer) {
        Timer--;
        if (Timer<=0) {
            Timer=0;
            TimerRunning=false;
            if (TimerMode) { // break timer was running
                TimerMode=false;
                playpause(); // playpause will start (or stop) the music
            } else {
                if (Soundfile!="") {
                    switch (PlayUsing) {
                    default:
                    case 0:
                        if (QSound::isAvailable()) {
                            QSound::play(Soundfile); // play tip timer sound using QSound (requires NAS under linux)
                        }
                        break;
                    case 1:
                        if (Command!="") {
                            QProcess::startDetached(Command.arg(Soundfile)); // play tip timer sound using command line (%1 gets replaced by Soundfile)
                        }
                        break;
                    }
                }
            }
            ui->pushButton_timer->setStyleSheet("background: red; border: 5px ridge #404000; font-size: 50px;"); // set red
        }
    }
    ui->pushButton_timer->setText(formatTime(Timer*1000)); // formatTime expects milliseconds
}

void MainWindow::deletePlaylistItems()
{
    int flag=0;

    while (ui->listView_playlist->selectionModel()->selectedIndexes().count()) {
        playlistModel->removeRows(ui->listView_playlist->selectionModel()->selectedIndexes().at(0).row(),1);
        flag=1;
    }
    if (flag) {
//        playlistChanged();
    }
}

void MainWindow::DoubleclickPlaylist(QModelIndex idx)
{
    nexttrack=idx.row();
    QString filename=playlistModel->getFilename(nexttrack);
    mediaObject->setCurrentSource(filename);
    mediaObject->play();
}

void MainWindow::metaDataChanged(void)
{
    QMultiMap<QString, QString> metaData = mediaObject->metaData();

#if 0
    ui->listWidget->clear();
    QMapIterator<QString, QString> i(metaData);
     while (i.hasNext()) {
         i.next();
         ui->listWidget->addItem(i.key()+"="+i.value());
     }
#endif
     ui->label_artist->setText(metaData.value("ARTIST"));
     ui->label_title->setText(metaData.value("TITLE"));
     ui->label_album->setText(metaData.value("ALBUM")); // id3v1 album not recognized!
     // use ID3v1 tags if necessary
     {
         char buf[128];
         char dat[31];
         qint64 size;
         QFile mp3file(mediaObject->currentSource().fileName());
         if (mp3file.open(QIODevice::ReadOnly)) {
             if ((size=mp3file.size())>=128) {
                 mp3file.seek(size-128);
                 mp3file.read(buf,128);
                 if (buf[0]=='T' && buf[1]=='A' && buf[2]=='G') { // valid ID3v1 tags
                     if (ui->label_title->text()=="") { // no title, try ID3v1 title
                         strncpy(dat,buf+3,30);
                         dat[30]=0;
                         ui->label_title->setText(dat);
                     }
                     if (ui->label_artist->text()=="") { // no artist, try ID3v1 artist
                         strncpy(dat,buf+33,30);
                         dat[30]=0;
                         ui->label_artist->setText(dat);
                     }
                     if (ui->label_album->text()=="") { // no album, try ID3v1 album
                         strncpy(dat,buf+63,30);
                         dat[30]=0;
                         ui->label_album->setText(dat);
                     }
                 }
             }
             mp3file.close();
         }
     }
}

void MainWindow::tick(qint64 time)
{
    qint64 ttime,reltime;

    reltime=time-zeroPosition();
    ttime=mediaObject->totalTime();
    ui->label_position->setText(formatTime(time));
    ui->label_length->setText(formatTime(ttime));
    ui->pushButton_counter->setText(formatTime(reltime));
}

void MainWindow::enqueueNext(void)
{
    QString filename;

    nexttrack=playlistModel->playPos()+1; // next track (default)
    if (ui->checkBox_loop->isChecked()) {
        if (Loopstart==0 && Loopend==0) {
            nexttrack--; // repeat same track
        } else {
            return; // looping portion of track
        }
    }
    if (nexttrack>=playlistModel->rowCount()) {
        nexttrack=-1; // no more tracks
    }
    if (nexttrack>=0) {
        filename=playlistModel->getFilename(nexttrack);
        mediaObject->clearQueue(); // this is in case it was already enqueued, but a seek back was done before current file finished playing
        mediaObject->enqueue(filename);
    }
}

void MainWindow::prefinishMark(void)
{
    if (ui->checkBox_loop->isChecked()) {
        mediaObject->seek(Loopstart);
        mediaObject->clearQueue(); // prefinish mark only works if queue is empty (just be sure)
        loopStateChanged(false); // clear the prefinish mark
        loopStateChanged(true); // set the prefinish mark again (necessary to make it work with gstreamer)
    }
}

void MainWindow::loopStateChanged(int state)
{
    qint64 msToEnd;

    if (state && Loopend>0) { // loop mode on
        msToEnd=mediaObject->totalTime()-Loopend;
        if (msToEnd<0) {
            msToEnd=0;
        }
        mediaObject->setPrefinishMark(msToEnd);
    } else { // loop mode off
        mediaObject->setPrefinishMark(0);
    }
}

void MainWindow::loop(void)
{
    ui->checkBox_loop->toggle();
}

// nexttrack was previously set to the next track to play, so it is the track playing now
void MainWindow::sourceChanged(void)
{
    struct position here;

    if (nexttrack>=0) { // do not auto-zero at start of track if nexttrack<0
        playlistModel->setPlayPos(nexttrack);
        here.track=nexttrack;
        here.position=0;
        markZero(here); // when changing tracks, mark zero at start of new track
    }
}

// new file playing, total time is known, look up loop times for this file
void MainWindow::totalTimeChanged(qint64 totalTime)
{
    QString filename;
    QFileInfo fileinfo;
    QFile loopfile;
    qint64 start,end;

    filename=mediaObject->currentSource().fileName();
    fileinfo.setFile(filename);
    filename=fileinfo.path()+"/"+fileinfo.completeBaseName()+".loop";
    loopfile.setFileName(filename);
    if (loopfile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QTextStream loop(&loopfile);
        start=loop.readLine().toLongLong();
        end=loop.readLine().toLongLong();
        loopfile.close();
        if (start<0) {
            start=0;
        }
        if (end<0) {
            end=0;
        }
        if (end<start) {
            end=start;
        }
    } else {
        start=end=0;
    }
    Loopstart=start;
    Loopend=end;
    loopStateChanged(ui->checkBox_loop->checkState()); // this will set proper prefinish time for looping
}

void MainWindow::stateChanged(Phonon::State newstate, Phonon::State oldstate)
{
    int playmode;

    switch (newstate) {
    case Phonon::PlayingState:
        if (Timer>0) {
            TimerRunning=true; // start timer (if it still has time left)
        }
        playmode=2;
        if (queuedFlag) { // perform the queued seek operation
            if (queuedPosition<0) { // negative means position from end of file
                queuedPosition=mediaObject->totalTime()+queuedPosition;
            }
            mediaObject->seek(queuedPosition);
            queuedFlag=false;
        }
        break;
    case Phonon::PausedState:
        playmode=1;
        if (queuedFlag) { // perform the queued seek operation
            if (queuedPosition<0) { // negative means position from end of file
                queuedPosition=mediaObject->totalTime()+queuedPosition;
            }
            mediaObject->seek(queuedPosition);
            queuedFlag=false; // no more queued seek
        }
        if (oldstate==Phonon::PlayingState) { // if it changed from playing to paused
            qint64 newposition=mediaObject->currentTime()-Backtime*1000; // back up
            if (newposition<0) {
                newposition=0;
            }
            mediaObject->seek(newposition);
        }
        break;
    case Phonon::StoppedState:
        ui->label_artist->setText("");
        ui->label_title->setText("");
        ui->label_album->setText("");
        ui->label_position->setText("");
        ui->label_length->setText("");
        ui->pushButton_counter->setText("0:00");
    default:
        playmode=0;
    }
    playlistModel->setPlayMode(playmode); // change playmode so playlist can display proper state
}

void MainWindow::dobreaktimer(void)
{
    Breaktimer dialog(this);
    static int time=50;

    dialog.setTime(time);
    if (dialog.exec()) {
        time=dialog.getTime();
        TimerMode=true; // set break timer mode
        Timer=time*6; // set timer in seconds (convert from tenth minutes)
        TimerRunning=true; // start the timer
        ui->pushButton_timer->setStyleSheet("background: yellow; border: 5px ridge #404000; font-size: 50px;"); // yellow backgroud is break timer indication
    }
}

void MainWindow::doloop(void)
{
    Loop dialog(this);
    QString filename;
    QFileInfo fileinfo;
    QFile loopfile;

    dialog.setTotalTime(mediaObject->totalTime());
    if (undoloc>=1) {
        dialog.zeroStart=undolist[undoloc-1].position;
    } else {
        dialog.zeroStart=0;
    }
    if (undoloc>=0) {
        dialog.zeroEnd=undolist[undoloc].position;
    } else {
        dialog.zeroEnd=0;
    }
    dialog.setLoopStart(Loopstart);
    dialog.setLoopEnd(Loopend);
    if (dialog.exec()) {
        Loopstart=dialog.getLoopStart();
        Loopend=dialog.getLoopEnd();
        loopStateChanged(ui->checkBox_loop->checkState()); // this will set proper prefinish time for looping
        // write data to the loop file
        filename=mediaObject->currentSource().fileName();
        fileinfo.setFile(filename);
        filename=fileinfo.path()+"/"+fileinfo.completeBaseName()+".loop";
        loopfile.setFileName(filename);
        if (loopfile.open(QIODevice::WriteOnly | QIODevice::Text)) {
            QTextStream loop(&loopfile);
            loop << Loopstart << "\n";
            loop << Loopend << "\n";
            loopfile.close();
        }
    }
}

void MainWindow::dosettings(void)
{
    settings dialog(this);

    dialog.setTiptime(Tiptime);
    dialog.setBacktime(Backtime);
    dialog.setSkiptime(Skiptime);
    dialog.setVolumestep(Volumestep);
    dialog.setSoundfile(Soundfile);
    dialog.setPlayUsing(PlayUsing);
    dialog.setCommand(Command);
    if (dialog.exec()) {
        Tiptime=dialog.getTiptime();
        Backtime=dialog.getBacktime();
        Skiptime=dialog.getSkiptime();
        Volumestep=dialog.getVolumestep();
        Soundfile=dialog.getSoundfile();
        PlayUsing=dialog.getPlayUsing();
        Command=dialog.getCommand();
        QSettings settings(QSettings::IniFormat,QSettings::UserScope,"krubow","sdplayer"); // constructor, force INI file storage
        settings.setValue("tiptime",Tiptime);
        settings.setValue("backtime",Backtime);
        settings.setValue("skiptime",Skiptime);
        settings.setValue("volumestep",Volumestep);
        settings.setValue("soundfile",Soundfile);
        settings.setValue("playusing",PlayUsing);
        settings.setValue("command",Command);
    }
}

void MainWindow::doshortcuts(void)
{
    int size,i;
    shortcuts dialog(this);
    struct shortcut newshortcut;

    size=shortcutlist.size();
    for (i=0; i<size; i++) {
        dialog.append(shortcutlist[i].Key, shortcutlist[i].Action);
    }
    if (dialog.exec()) {
        deactivateShortcuts();
        shortcutlist.clear();
        size=dialog.shortcutsCount();
        for (i=0; i<size; i++) {
            newshortcut.Key=dialog.getKey(i);
            newshortcut.Action=dialog.getAction(i);
            shortcutlist.append(newshortcut);
        }
        activateShortcuts();
        QSettings settings(QSettings::IniFormat,QSettings::UserScope,"krubow","sdplayer"); // constructor, force INI file storage
        settings.beginWriteArray("shortcuts");
        size=shortcutlist.size();
        for (i=0; i<size; i++) {
            settings.setArrayIndex(i);
            settings.setValue("key",shortcutlist[i].Key);
            settings.setValue("action",shortcutlist[i].Action);
        }
        settings.endArray();
    }
}

void MainWindow::activateShortcuts(void)
{
    int i;

    for (i=0; i<shortcutlist.size(); i++) {  // create shortcuts
          connect(shortcutlist[i].ptrShortcut= new QShortcut(shortcutlist[i].Key,this),SIGNAL(activated()),this,shortcutSlots.value(shortcutlist[i].Action));
    }
}

void MainWindow::deactivateShortcuts(void)
{
    int i;

    for (i=0; i<shortcutlist.size(); i++) { // destroy all existing shortcuts
          delete shortcutlist[i].ptrShortcut;
    }
}

// pause if playing, else play
void MainWindow::playpause(void)
{
    if (playlistModel->playPos()<0) { // no track playing
        next(); // attempt to play next (i.e. first) track
    } else {
        if (mediaObject->state()==Phonon::PlayingState) {
            mediaObject->pause();
        } else {
            mediaObject->play();
        }
    }
}

// back Skiptime seconds
// it is complicate by allowing backing up across files, whether playing or paused
void MainWindow::back(void)
{
    qint64 newtime;
    bool wasplaying;

    wasplaying=false;
    switch (mediaObject->state()) {
    case Phonon::PlayingState:
        wasplaying=true; // set flag and fall thru to next case
    case Phonon::PausedState:
        newtime=mediaObject->currentTime()-Skiptime*1000; // back up by Skiptime seconds
        if (newtime<0) { // possible back to prev track
            if (playlistModel->playPos()>0) { // if there is a previous track
                nexttrack=playlistModel->playPos()-1;
                mediaObject->setCurrentSource(playlistModel->getFilename(nexttrack));
                if (wasplaying) {
                    mediaObject->play();
                } else {
                    mediaObject->pause();
                }
                queuedPosition=newtime; // queue the seek to be done later
                queuedFlag=true;
                break;
            }
            newtime=0; // no prev track, limit back to start of track
        }
        mediaObject->seek(newtime); // change position in current track
        break;
    default:
        break;
    }
}

// forward Skiptime seconds
void MainWindow::forward(void)
{
    qint64 newtime;
    bool wasplaying;

    wasplaying=false;
    switch (mediaObject->state()) {
    case Phonon::PlayingState:
        wasplaying=true; // set flag and fall thru to next case
    case Phonon::PausedState:
        newtime=mediaObject->currentTime()+Skiptime*1000; // forward by Skiptime seconds
        if (newtime>mediaObject->totalTime()) { // possible forward to next track
            if (playlistModel->playPos()+1 < playlistModel->rowCount()) { // if there is a next track
                nexttrack=playlistModel->playPos()+1;
                newtime=newtime-mediaObject->totalTime();
                mediaObject->setCurrentSource(playlistModel->getFilename(nexttrack));
                if (wasplaying) {
                    mediaObject->play();
                } else {
                    mediaObject->pause();
                }
                queuedPosition=newtime; // queue the seek to be done later
                queuedFlag=true;
                break;
            }
            newtime=mediaObject->totalTime(); // no next track, limit forward to end of track
        }
        mediaObject->seek(newtime);
        break;
    default:
        break;
    }
}

/*
The undo list contains a list of positions (track,position)
undoloc is the position of the current zero mark on the list (-1 if none)
The list can have entries with an index>undoloc if undo's have been done.
In this case, a redo is possible, incrementing undoloc up to a max of size-1
*/
void MainWindow::emptyUndo(void)
{
    undoloc=-1;
    undolist.clear();
}

int MainWindow::zeroTrack() // return track of current zero mark, -1 if none
{
    if (undoloc<0) {
        return -1;
    }
    return undolist[undoloc].track;
}

qint64 MainWindow::zeroPosition() // return position of current zero mark, -1 if none
{
    if (undoloc<0) {
        return -1;
    }
    return undolist[undoloc].position;
}

// enter new position in undo list if different from current zero position
void MainWindow::markZero(struct position here)
{
    if (here.track!=zeroTrack() || here.position!=zeroPosition()) {
        undoloc++;
        if (undolist.size()>undoloc) { // already room in undolist
            undolist[undoloc]=here;
            while (undolist.size()>undoloc+1) {
                undolist.removeLast();
            }
        } else {
            undolist.append(here);
        }
        tick(mediaObject->currentTime()); // update counters
    }
}

// enter current position in undo list
void MainWindow::zero(void)
{
    struct position here;

    here.track=playlistModel->playPos(); // get current position
    here.position=mediaObject->currentTime();
    markZero(here);
}

// seek to current entry in undo list
void MainWindow::rewind(void)
{
    int ztrack;
    qint64 zposition;
    QString filename;
    bool wasplaying;

    ztrack=zeroTrack();
    zposition=zeroPosition();
    if (playlistModel->playPos()!=ztrack) { // need to change track
        wasplaying=(mediaObject->state()==Phonon::PlayingState);
        nexttrack=ztrack;
        filename=playlistModel->getFilename(nexttrack);
        mediaObject->setCurrentSource(filename);
        if (wasplaying) {
            mediaObject->play();
        } else {
            mediaObject->pause();
        }
        queuedPosition=zposition; // queue the seek to be done later
        queuedFlag=true;
    } else { // simple case, seek in current track
        mediaObject->seek(zposition);
    }
}

void MainWindow::tiptimer(void)
{
    TimerMode=false; // cancel break timer mode
    if (TimerRunning) { // if running, stop the timer
        TimerRunning=false;
    } else {  // else reset the timer and set background to white
        Timer=Tiptime*60;
        ui->pushButton_timer->setStyleSheet("background: white; border: 5px ridge #404000; font-size: 50px;"); // clear red
        if (mediaObject->state()==Phonon::PlayingState) { // if playing, start the timer
            TimerRunning=true;
        }
        ui->pushButton_timer->setText(formatTime(Timer*1000)); // formatTime expects milliseconds
    }
}

void MainWindow::next(void)
{
    QString filename;

    nexttrack=playlistModel->playPos()+1;
    if (nexttrack>=playlistModel->rowCount()) {
        nexttrack-=1; // no more tracks
    } else {
        filename=playlistModel->getFilename(nexttrack);
        mediaObject->setCurrentSource(filename);
        mediaObject->play();
    }
}

void MainWindow::prev(void)
{
    QString filename;

    nexttrack=playlistModel->playPos()-1;
    if (nexttrack<0) {
        nexttrack+=1; // no prev tracks
    } else {
        filename=playlistModel->getFilename(nexttrack);
        mediaObject->setCurrentSource(filename);
        mediaObject->play();
    }
}

void MainWindow::volup(void)
{
    qreal volume;

    volume=audioOutput->volume()+((qreal)Volumestep)/100;
    if (volume>1.0) {
        volume=1.0;
    }
    audioOutput->setVolume(volume);
}

void MainWindow::voldown(void)
{
    qreal volume;

    volume=audioOutput->volume()-((qreal)Volumestep)/100;
    if (volume<0) {
        volume=0;
    }
    audioOutput->setVolume(volume);
}

void MainWindow::undo(void)
{
    if (undoloc>0) {
        undoloc--;
        rewind();
    }
}

void MainWindow::redo(void)
{
    if (undoloc+1<undolist.size()) {
        undoloc++;
        rewind();
    }
}

void MainWindow::playlistChanged(void)
{
    PlaylistChanged=true;
}

void MainWindow::setTitle(void)
{
    setWindowTitle(QString("SDplayer ")+QFileInfo(Openfile).fileName()); // display filename without path
}

void MainWindow::saveFile(QString filename)
{
    QFileInfo fileinfo;
    QFile playlistfile;
    QString playlistpath;
    int pathlength;
    QString mp3filename;
    int track;
    qint64 pos;

    fileinfo.setFile(filename);
    if (fileinfo.suffix().toLower()!="m3u") {   // if not a "m3u" filename
        filename+=".m3u"; // append .m3u file type
    }
    playlistpath=fileinfo.path()+"/"; // get path, complete with terminating '/'
    pathlength=playlistpath.size();
    playlistfile.setFileName(filename);
    if (playlistfile.open(QIODevice::WriteOnly | QIODevice::Text)) {
        QTextStream m3u(&playlistfile);
        track=zeroTrack();
        if (track<0) {
            track=0;
        }
        pos=zeroPosition();
        if (pos<0) {
            pos=0;
        }
        m3u << "#TRACK:" << track << "\n";
        m3u << "#POSITION:" << pos << "\n";
        for (int row=0; row<playlistModel->rowCount(); row++) {
            mp3filename=playlistModel->getFilename(row);
            // convert file path to path relative to playlist if possible
            if (mp3filename.startsWith(playlistpath)) {
                mp3filename.remove(0,pathlength);
            }
            m3u << mp3filename << "\n";
        }
        playlistfile.close();
    }
}

void MainWindow::saveAs(void)
{
    QString filename;
    static QString mydir=QDir::homePath(); // last visited directory

    filename=QFileDialog::getSaveFileName(this,"Save Playlist File",mydir,"M3U file (*.m3u)");
    if (filename!="") {
        saveFile(filename);
        Openfile=filename;
        setTitle();
        QDir path(Openfile);
        mydir=path.absolutePath(); // remember last visited directory
        PlaylistChanged=false;
    }
}

void MainWindow::save(void)
{
    if (Openfile=="") {
        saveAs();
    } else {
        saveFile(Openfile);
        PlaylistChanged=false;
    }
}

void MainWindow::closefile(void)
{
    saveAsk();
    if (playlistModel->rowCount()>0) {
        playlistModel->removeRows(0,playlistModel->rowCount());
    }
    mediaObject->stop(); // don't leave a file playing when I close the playlist
    Openfile="";
    setTitle();
    emptyUndo();
    PlaylistChanged=false;
}

void MainWindow::saveAsk(void)
{
    QString message;

    if (PlaylistChanged || undoloc>0) { // playlist changed or additional zero point is set
        message=QString("Save changes to playlist?\n%1").arg(Openfile);
        if (QMessageBox::Save==QMessageBox::question(this,"Save File?",message,QMessageBox::Save|QMessageBox::Discard,QMessageBox::Save)) {
            save();
        }
    }
}

void MainWindow::openPlaylist(QString filename)
{
    QFileInfo fileinfo;
    QFile playlistfile;
    QString playlistpath;
    int row;
    QString line;
    int track;
    qint64 position;
    struct position here;

    if (filename=="") {
        return; // do nothing if there is no file to open
    }
    closefile(); // must close any open playlist before opening a new one (will prompt to save)
    fileinfo.setFile(filename);
    if (fileinfo.suffix().toLower()!="m3u") {   // if not a "m3u" filename
        QMessageBox::information(this,"title","Appending m3u filetype");
        filename+=".m3u"; // append .m3u file type
    }
    playlistpath=fileinfo.path()+"/"; // path of playlist file for converting relative mp3 files back to absolute
    playlistfile.setFileName(filename);
    if (playlistfile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QTextStream m3u(&playlistfile);
        row=0;
        track=0;
        position=0;
        while (!m3u.atEnd()) {
            line=m3u.readLine();
            if (line[0]!='#') {
                fileinfo.setFile(line);
                if (fileinfo.isRelative()) { // if mp3 file has relative path
                    line.prepend(playlistpath); // convert it back to absolute path
                }
                playlistModel->insertRow(row);
                playlistModel->setData(playlistModel->index(row,0,QModelIndex()),line);
                row++;
            } else {
                if (line.startsWith("#TRACK:")) {
                    track=line.remove(0,7).toInt();
                } else {
                    if (line.startsWith("#POSITION:")) {
                        position=line.remove(0,10).toLongLong();
                    }
                }
            }
        }
        playlistfile.close();
        Openfile=filename;
        setTitle();
        emptyUndo();
        here.track=track;
        here.position=position;
        markZero(here);
        rewind(); // this seeks to the zero location I marked, but marks a new zero at start of new track
        emptyUndo(); // so I get rid of the unwanted zero
        markZero(here); // and mark the zero location again so it is the only zero marked
        PlaylistChanged=false;
    }
}

void MainWindow::open(void)
{
    QString filename;
    static QString mydir=QDir::homePath(); // last visited directory

    filename=QFileDialog::getOpenFileName(this,"Open Playlist File",mydir,"M3U file (*.m3u)");
    if (filename!="") {
        openPlaylist(filename);
        QDir path(filename);
        mydir=path.absolutePath(); // remember last visited directory
    }
}

void MainWindow::dragEnterEvent(QDragEnterEvent *event)
{
  if (event->mimeData()->hasUrls()) { // if dragged item has list of urls (filenames)
    if (event->mimeData()->urls().size()==1) { // if list contains only a single filename
      event->acceptProposedAction();
    }
  }
}

void MainWindow::dropEvent(QDropEvent *event)
{
  QString filename;
  QFileInfo fileinfo;

  if (event->mimeData()->hasUrls()) {
    foreach (QUrl url, event->mimeData()->urls()) {
      filename=url.toLocalFile();
      fileinfo.setFile(filename); // get information about file
      if (fileinfo.isFile()) { // if it is a file (not a directory or symlink)
        if (fileinfo.suffix().toLower()=="m3u") { // playlist file
//          SaveAsk();
//          Openfile=filename; // dropped mp3 file
//          OpenGrpFile();
//            QMessageBox::information(this,"title","Dropped an m3u file!");
            openPlaylist(filename);
        }
      }
    }
  }
}

void MainWindow::playlistDropped(QString filename)
{
    openPlaylist(filename);
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    saveAsk();
    QSettings settings(QSettings::IniFormat,QSettings::UserScope,"krubow","sdplayer"); // constructor, force INI file storage
    settings.setValue("volume",audioOutput->volume()); // save volume setting
    settings.setValue("openfile",Openfile); // save the open playlist file name
    event->accept();
}
